# AUTOGENERATED BY NBDEV! DO NOT EDIT!

__all__ = ["index", "modules", "custom_doc_links", "git_url"]

index = {"ode_solve": "00_ode.ipynb",
         "ODEF": "00_ode.ipynb",
         "ODEAdjoint": "00_ode.ipynb",
         "NeuralODE": "00_ode.ipynb",
         "MMD_loss": "01_losses.ipynb",
         "OT_loss": "01_losses.ipynb",
         "Density_loss": "01_losses.ipynb",
         "Local_density_loss": "01_losses.ipynb",
         "group_extract": "02_utils.ipynb",
         "sample": "02_utils.ipynb",
         "to_np": "02_utils.ipynb",
         "generate_steps": "02_utils.ipynb",
         "set_seeds": "02_utils.ipynb",
         "config_hold_out": "02_utils.ipynb",
         "config_criterion": "02_utils.ipynb",
         "get_groups_from_df": "02_utils.ipynb",
         "get_cell_types_from_df": "02_utils.ipynb",
         "get_sample_n_from_df": "02_utils.ipynb",
         "get_times_from_groups": "02_utils.ipynb",
         "ToyODE": "03_models.ipynb",
         "make_model": "03_models.ipynb",
         "Autoencoder": "03_models.ipynb",
         "ToyModel": "03_models.ipynb",
         "ToySDEModel": "03_models.ipynb",
         "plot_losses": "04_plots.ipynb",
         "plot_comparision": "04_plots.ipynb",
         "new_plot_comparisions": "04_plots.ipynb",
         "plot_gene_trends": "04_plots.ipynb",
         "train": "05_train.ipynb",
         "train_ae": "05_train.ipynb",
         "training_regimen": "05_train.ipynb",
         "__file": "06_constants.ipynb",
         "ROOT_DIR": "06_constants.ipynb",
         "DATA_DIR": "06_constants.ipynb",
         "NTBK_DIR": "06_constants.ipynb",
         "IMGS_DIR": "06_constants.ipynb",
         "RES_DIR": "06_constants.ipynb",
         "WORM_FILE": "06_constants.ipynb",
         "EB_BODIES_FILE": "06_constants.ipynb",
         "EB_BODIES_PSEUDO_4": "06_constants.ipynb",
         "EB_BODIES_PSEUDO_6": "06_constants.ipynb",
         "EB_BODIES_PSEUDO_25": "06_constants.ipynb",
         "EB_BODIES_PSEUDO_82": "06_constants.ipynb",
         "DYNGEN_INFO_FILE": "06_constants.ipynb",
         "DYNGEN_EXPR_FILE": "06_constants.ipynb",
         "construct_diamond": "07_datasets.ipynb",
         "make_diamonds": "07_datasets.ipynb",
         "make_swiss_roll": "07_datasets.ipynb",
         "make_tree": "07_datasets.ipynb",
         "make_worm_data": "07_datasets.ipynb",
         "make_eb_data": "07_datasets.ipynb",
         "make_dyngen_data": "07_datasets.ipynb",
         "relabel_data": "07_datasets.ipynb",
         "rings": "07_datasets.ipynb",
         "make_rings": "07_datasets.ipynb",
         "make_jacks": "07_datasets.ipynb",
         "config_exp_logger": "08_exp.ipynb",
         "exp_log_filename": "08_exp.ipynb",
         "exp_param_filename": "08_exp.ipynb",
         "list_exps": "08_exp.ipynb",
         "gen_exp_name": "08_exp.ipynb",
         "load_exp_params": "08_exp.ipynb",
         "save_exp_params": "08_exp.ipynb",
         "setup_exp": "08_exp.ipynb",
         "is_config_subset": "08_exp.ipynb",
         "find_exps": "08_exp.ipynb",
         "test_exp": "08_exp.ipynb",
         "DiffusionDistance": "09_geo.ipynb",
         "DiffusionAffinity": "09_geo.ipynb",
         "DiffusionMap": "09_geo.ipynb",
         "PhateDistance": "09_geo.ipynb",
         "old_DiffusionDistance": "09_geo.ipynb",
         "setup_distance": "09_geo.ipynb",
         "generate_points": "10_eval.ipynb",
         "generate_trajectories": "10_eval.ipynb",
         "generate_plot_data": "10_eval.ipynb",
         "get_points_from_trajectories": "10_eval.ipynb",
         "calculate_nn": "10_eval.ipynb",
         "generate_tjnet_trajectories": "10_eval.ipynb",
         "get_cell_indexes": "10_eval.ipynb"}

modules = ["ode.py",
           "losses.py",
           "utils.py",
           "models.py",
           "plots.py",
           "train.py",
           "constants.py",
           "datasets.py",
           "exp.py",
           "geo.py",
           "eval.py"]

doc_url = "https://KrishnaswamyLab.github.io/MIOFlow/"

git_url = "https://github.com/KrishnaswamyLab/MIOFlow/tree/main/"

def custom_doc_links(name): return None
